/*
 * (C) Copyright 2005- ECMWF.
 *
 * This software is licensed under the terms of the Apache Licence Version 2.0
 * which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
 *
 * In applying this licence, ECMWF does not waive the privileges and immunities granted to it by
 * virtue of its status as an intergovernmental organisation nor does it submit to any jurisdiction.
 */

#ifndef eccodes_version_H
#define eccodes_version_H

#define ECCODES_VERSION_STR      "2.44.0"

#define ECCODES_MAJOR_VERSION    2
#define ECCODES_MINOR_VERSION    44
#define ECCODES_REVISION_VERSION 0
#define ECCODES_VERSION (ECCODES_MAJOR_VERSION*10000+ECCODES_MINOR_VERSION*100+ECCODES_REVISION_VERSION)

#endif /* eccodes_version_H */
